#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "lib/GC.h"
#include "apack_streams.h"

t_stream create_stream(const char *file,int flags,int bsize){
 t_stream new_stream; memsetb(&new_stream,0,sizeof(t_stream));
 new_stream.fp=open(file,flags);
 new_stream.bsize=bsize;
 new_stream.buffer=(char*)malloc(bsize);
 memsetb(new_stream.buffer,0,bsize);
 return new_stream;
}

void close_stream(t_stream *s){
 close(s->fp);
 free(s->buffer);
 memsetb(s,0,sizeof(t_stream));
}

int get_bit(t_stream *s,int *bit){
 if(!s->bytes_left)s->bytes_left=read(s->fp,s->buffer,s->bsize);
 if(s->bit_cursor<8){
  *bit=(s->buffer[s->byte_cursor]>>(7-s->bit_cursor))&1;
  s->bit_cursor++;
 }else{
   s->bit_cursor=0;
   s->byte_cursor++;
   if(s->byte_cursor>=s->bsize){
    s->byte_cursor=0;
    if(!(s->bytes_left=read(s->fp,s->buffer,s->bsize)))return 0;
   }
   if(s->byte_cursor>=s->bytes_left)return 0;
   *bit=(s->buffer[s->byte_cursor]>>7)&1;
   s->bit_cursor=1;
  }
 return -1;
}

int get_byte(t_stream *s,unsigned char *byte){
 if(!s->bytes_left)s->bytes_left=read(s->fp,s->buffer,s->bsize);
 if(s->byte_cursor>=s->bytes_left){
  if(s->bytes_left<s->bsize)return 0;
  s->bytes_left=read(s->fp,s->buffer,s->bsize);
  s->byte_cursor=0;
 }

 *byte=s->buffer[s->byte_cursor];
 s->byte_cursor++;
 s->bit_cursor+=8;
 s->bit_cursor&=7;

 return -1;
}

void reset_stream(t_stream *s){
 lseek(s->fp,0,SEEK_SET);
 s->byte_cursor=0;
 s->bit_cursor=0;
 s->bytes_left=read(s->fp,s->buffer,s->bsize);
}

void sync_stream(t_stream *s){
 s->byte_cursor=0;
 s->bit_cursor=0;
 s->bytes_left=read(s->fp,s->buffer,s->bsize);
}

void write_byte(t_stream *s,unsigned char byte){
 if(s->bytes_left<s->bsize){
  s->buffer[s->bytes_left]=byte;
  s->bytes_left++;
 }else{while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left));
       s->buffer[0]=byte; s->bytes_left=1;
      }
}

void flush_write(t_stream *s){s->bytes_left+=(s->bit_cursor)?1:0; while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left)); memsetb(s->buffer,0,s->bsize); s->bit_cursor=0;}

void write_bit(t_stream *s,char bit){
 if(s->bit_cursor==8){s->bit_cursor=0; s->bytes_left++;}
 if(s->bytes_left<s->bsize){
  s->buffer[s->bytes_left]|=bit<<(7-s->bit_cursor);
  s->bit_cursor++;
 }else{while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left));
       memsetb(s->buffer,0,s->bsize);
       s->buffer[0]=bit<<7; s->bit_cursor=1;
      }
}

int get_file_stats_by_byte(t_stream *s,t_node *node){
 int i;for(i=0;i<512;i++)memsetb(&node[i],0,sizeof(t_node));
 for(i=0;i<256;i++)node[i].val=i;
 for(i=256;i<512;i++)node[i].val=-1;
 int byte=0;
 while(get_byte(s,(unsigned char*)&byte)){node[byte].frq++; byte=0;}
 MLQuickSortl(node,0,sizeof(t_node),0,255);
 i=0; while((node[i].frq)&&(i<256))i++;
 LMQuickSortl(node,0,sizeof(t_node),0,i-1);
 return i;
}

void build_index(t_node *in,t_node *out,int len){int i;for(i=0;i<len;i++){in[i].l_leaf=NULL; in[i].r_leaf=NULL; in[i].root=NULL; out[in[i].val]=in[i]; in[i].root=&out[in[i].val];}}

void build_tree(t_node *node, t_node *index,int *len){
 if((*len)<1)return;
 t_node *enode; //Manejadores para poder actualizar los punteros.
 int from=0,to=*len-1,i_nodes=256;//*len;
 
 if(to)do{
  LMQuickSortl(node,0,sizeof(t_node),from,to);
  to++;
  node[to].val=-1; node[to].super=NULL;
  node[to].frq=node[from].frq+node[from+1].frq;
  index[i_nodes]=node[to];
  node[to].root=&index[i_nodes];

  enode=node[from].root;
  enode->super=&index[i_nodes];
  enode=node[from+1].root;
  enode->super=&index[i_nodes];

  enode=node[from].root;
  enode->bit=1;
  index[i_nodes].l_leaf=enode;

  enode=node[from+1].root;
  enode->bit=0;
  index[i_nodes].r_leaf=enode;

  i_nodes++;
  from+=2;
 }while(from<to);else{node[++to].val=-1; node[to].super=NULL;
                      node[to].frq=node[from].frq;
                      index[i_nodes]=node[to];
                      node[to].root=&index[i_nodes];

                      enode=node[from].root;
                      enode->super=&index[i_nodes];

                      enode=node[from].root;
                      enode->bit=1;
                      index[i_nodes].l_leaf=enode;
                     }

 *len=to+1; //No importa.
}

void write_huffman_code(const t_node *node,t_stream *s){
 if(!node->super)return;else write_huffman_code(node->super,s);
 write_bit(s,node->bit);
}

void print_huffman_code(const t_node *node){
 if(!node->super)return;else print_huffman_code(node->super);
 printf("%d",node->bit);
}

int huffman2byte(t_node *root,t_node **cur,int bit,int *byte){
 *cur=(!(*cur))?root:*cur;
 *cur=(bit)?(*cur)->l_leaf:(*cur)->r_leaf; if(!(*cur))return 0;
 if(((long)((*cur)->l_leaf))|((long)((*cur)->r_leaf)))return 0;
 *byte=(*cur)->val;
 *cur=NULL;
 return -1;
}

t_node *get_root(t_node *n){while(n->super)n=n->super; return n;}

#define BUFFER_BUFFERED 1024

#ifndef DECOMP
int main(int argc,char **argv){
 t_stream sin=create_stream(argv[1],O_RDONLY,BUFFER_BUFFERED),
          sout=create_stream(argv[2],O_WRONLY|O_CREAT,BUFFER_BUFFERED);
 t_node node[512],index[512];
 int nodes=get_file_stats_by_byte(&sin,node);
 printf("Nodos identificados: %d\n",nodes);

 write(sout.fp,"HUFF ",5);
 int i=nodes-1;
 write(sout.fp,&i,1); //Posicion 4 reservada para el byte de fragmentacion
 for(i=0;i<nodes;i++){write(sout.fp,&node[i].val,1); write(sout.fp,&node[i].frq,4);}
 build_index(node,index,nodes);
 build_tree(node,index,&nodes);
 printf("Nodos ramificados: %d\n",nodes);
 reset_stream(&sin);
 unsigned char byte,fragmentation=0;
 int progress=0; while(get_byte(&sin,&byte)){fragmentation++; fragmentation&=7; if(++progress>1024){putchar('.');progress=0;} write_huffman_code(&index[byte],&sout); byte=0;}; flush_write(&sout);
                 lseek(sout.fp,4,SEEK_SET); write(sout.fp,&fragmentation,1);
 printf("Ok.\n");
 close_stream(&sin);
 close_stream(&sout);
 return 0;
}
#else
int main(int argc,char **argv){
 t_stream sin=create_stream(argv[1],O_RDONLY,BUFFER_BUFFERED),
          sout=create_stream(argv[2],O_WRONLY|O_CREAT,BUFFER_BUFFERED);
 t_node node[512],index[512];
 char signature[5]={'\0'};
 read(sin.fp,signature,4);
 int nodes=0,ifrag=0,total_size=0;
 read(sin.fp,&ifrag,1); ifrag=(ifrag)?1:0;
 read(sin.fp,&nodes,1); nodes++;
 printf("Nodos identificados: %d\n",nodes);
 int i;for(i=0;i<nodes;i++){node[i].val=0; read(sin.fp,&node[i].val,1); node[i].frq=0; read(sin.fp,&node[i].frq,4); total_size+=node[i].frq;}

 build_index(node,index,nodes);
 build_tree(node,index,&nodes);
 printf("Nodos ramificados: %d\n",nodes);
 i=node[0].val;
 t_node *cursor=NULL,*root=get_root(&index[i]);
 int bit=0,byte=0,progress=0,bytes_written=0;
 while((get_bit(&sin,&bit))&&(bytes_written<total_size)){if(++progress>1024){putchar('.');progress=0;} if(huffman2byte(root,&cursor,bit,&byte)){write_byte(&sout,byte); byte=0; bytes_written++;} bit=0;}; flush_write(&sout);
 printf("Ok.\n");
 close_stream(&sin);
 close_stream(&sout);
 return 0;
}
#endif

