#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "lib/GC.h"
#include "apack_streams.h"
long tmp_s=0;


t_stream create_stream(const char *file,int flags,int bsize){
 t_stream new_stream; memsetb(&new_stream,0,sizeof(t_stream));
 new_stream.fp=open(file,flags);
 new_stream.bsize=bsize;
 new_stream.buffer=(char*)malloc(bsize);
 memsetb(new_stream.buffer,0,bsize);
 return new_stream;
}

void close_stream(t_stream *s){
 close(s->fp);
 free(s->buffer);
 memsetb(s,0,sizeof(t_stream));
}

int get_bit(t_stream *s,int *bit){
 if(!s->bytes_left)s->bytes_left=read(s->fp,s->buffer,s->bsize);
 if(s->bit_cursor<8){
  *bit=(s->buffer[s->byte_cursor]>>(7-s->bit_cursor))&1;
  s->bit_cursor++;
 }else{
   s->bit_cursor=0;
   s->byte_cursor++;
   if(s->byte_cursor>=s->bsize){
    s->byte_cursor=0;
    if(!(s->bytes_left=read(s->fp,s->buffer,s->bsize)))return 0;
   }
   if(s->byte_cursor>=s->bytes_left)return 0;
   *bit=(s->buffer[s->byte_cursor]>>7)&1;
   s->bit_cursor=1;
  }
 return -1;
}

int get_byte(t_stream *s,unsigned char *byte){
 if(!s->bytes_left)if(!(s->bytes_left=read(s->fp,s->buffer,s->bsize)))return 0;
 if(s->byte_cursor>=s->bytes_left){
  //if(s->bytes_left<s->bsize)return s->bytes_left=0;
  s->byte_cursor=0;
  if(!(s->bytes_left=read(s->fp,s->buffer,s->bsize)))return 0;
 }
tmp_s=(s->buffer[s->byte_cursor]==0xF3)?-1:0;
 *byte=s->buffer[s->byte_cursor];
 s->byte_cursor++;
 s->bit_cursor+=8;
 s->bit_cursor&=7;

 return -1;
}

int get_word(t_stream *s,unsigned short *word){
 int msb=0,lsb=0;
 *word=0;
 if(!get_byte(s,(unsigned char*)&lsb))return 0;
 if(!get_byte(s,(unsigned char*)&msb))return 0;
 *word=(unsigned short)((msb<<8)|lsb);
 return -1;
}

int get_bytes(t_stream *s,char *byte,int bytes){
 int i,err=-1;for(i=0;i<bytes;i++)err&=get_byte(s,&byte[i]);
 return err;
}

void reset_stream(t_stream *s){
 lseek(s->fp,0,SEEK_SET);
 s->byte_cursor=0;
 s->bit_cursor=0;
 s->bytes_left=read(s->fp,s->buffer,s->bsize);
}

void sync_stream(t_stream *s){
 s->byte_cursor=0;
 s->bit_cursor=0;
 s->bytes_left=read(s->fp,s->buffer,s->bsize);
}

void write_byte(t_stream *s,unsigned char byte){
 if(s->bytes_left<s->bsize){
  s->buffer[s->bytes_left]=byte;
  s->bytes_left++;
 }else{while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left));
       memsetb(s->buffer,0,s->bsize);
       s->buffer[0]=byte; s->bytes_left=1;
      }
 s->bit_cursor=0;
}

void write_bytes(t_stream *s,void *_byte,int bytes){
 int i;for(i=0;i<bytes;i++)write_byte(s,((char*)_byte)[i]);
/* char *byte=(char*)_byte;
 if((bytes>=s->bsize)||((s->bytes_left+bytes)>=s->bsize)){
  while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left));
  write(s->fp,byte,bytes); return;
 }
 memcpyb(&s->buffer[s->bytes_left],byte,bytes);
 s->bytes_left+=bytes;*/
}

void flush_write(t_stream *s){s->bytes_left+=(s->bit_cursor)?1:0; while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left)); memsetb(s->buffer,0,s->bsize); s->bit_cursor=0;}

void write_bit(t_stream *s,char bit){
 if(s->bit_cursor==8){s->bit_cursor=0; s->bytes_left++;}
 if(s->bytes_left<s->bsize){
  s->buffer[s->bytes_left]|=bit<<(7-s->bit_cursor);
  s->bit_cursor++;
 }else{while(s->bytes_left-=write(s->fp,s->buffer,s->bytes_left));
       memsetb(s->buffer,0,s->bsize);
       s->buffer[0]=bit<<7; s->bit_cursor=1;
      }
}

long in_dict(t_ANSIstring *dict,int dict_sz,char *acum,int acum_sz){
 if((!dict)||(!acum_sz))return -1;
 int i=0,keep_doing=-1,no_match_found=-1;
 do{
  if(dict[i].len==acum_sz){
   keep_doing=memcompb(dict[i].bytes,acum,acum_sz);
   no_match_found=keep_doing;
  }
  keep_doing=(++i<dict_sz)?keep_doing:0;
 }while(keep_doing);

 return (no_match_found)?-1:--i;
}

#define SYSTEM_BUS_WIDE 8

void lz_compress(t_stream *in,t_stream *out){
 unsigned long grows_at=0xFFFF,bytes_count=2;
 long acumulator_sz=2,ai=1,di=0,dict_sz=0,entry;
 unsigned char *acumulator=NULL;
 t_ANSIstring *dict=NULL;

 acumulator=(char*)malloc(2); acumulator[0]=0; acumulator[1]=0;
 int read_stat=get_byte(in,&acumulator[0]); if(!read_stat){write_bytes(out,acumulator,bytes_count); free(acumulator); return;}

 do{
  if(read_stat=get_byte(in,&acumulator[ai])){
   if((entry=in_dict(dict,di,acumulator,ai+1))==-1){
    if(di<grows_at){
     if(di>=dict_sz)dict=(t_ANSIstring*)realloc(dict,sizeof(t_ANSIstring)*(++dict_sz));
     dict[di].bytes=malloc(ai+1); memcpy(dict[di].bytes,acumulator,ai+1); dict[di].len=ai+1; di++;
     if((entry=in_dict(dict,di,acumulator,ai))==-1)entry=(int)acumulator[0];else entry+=256;
     write_bytes(out,&entry,bytes_count);
     acumulator[0]=acumulator[ai];
     ai=1;
    }else{if((entry=in_dict(dict,di,acumulator,ai))==-1)entry=(int)acumulator[0];else entry+=256;
          write_bytes(out,&entry,bytes_count);
          acumulator[0]=acumulator[ai];
          ai=1;
#ifndef DINAMIC_DIC_BITS
          di=0;
#else
#warning Posiblemente NO mejore los resultados.
          entry=grows_at; write_bytes(out,&entry,bytes_count); bytes_count++; grows_at<<=8; grows_at|=0xFF;
          if(bytes_count>SYSTEM_BUS_WIDE){di=0; bytes_count=2; grows_at=0xFFFF;}
#endif
         }
   }else{
     if(ai>=acumulator_sz){
      acumulator_sz++;
      acumulator=(char*)realloc(acumulator,acumulator_sz+1);
     }
     ai++;
    }
  }
 }while(read_stat);

 if(ai)if((entry=in_dict(dict,di,acumulator,ai+1))==-1){
        if((entry=in_dict(dict,di,acumulator,ai))==-1)entry=(int)acumulator[0];else entry+=256;
        write_bytes(out,&entry,bytes_count);
       }else{entry+=256; write_bytes(out,&entry,bytes_count);}
 flush_write(out);
 if(acumulator)free(acumulator);
 if(dict)free(dict);
}


void lz_expand(t_stream *in,t_stream *out){
 unsigned long grows_at=0xFFFF,bytes_count=2;
 unsigned char *acumulator=NULL,byte;
 unsigned short word;
 long acumulator_sz=0,ai=-1,di=-1,dict_sz=0,entry;
 t_ANSIstring *dict=NULL;

 while(get_word(in,&word)){
  if(word==0xFFFF){di=-1; continue;}

  if(ai>=0){
   if((++ai)>=acumulator_sz)acumulator=(unsigned char*)realloc(acumulator,++acumulator_sz);
   if(word<256){
    byte=(unsigned char)(word&0xFF);
    write_byte(out,byte);

    acumulator[ai]=byte;
    if((++di)>=dict_sz)dict=(t_ANSIstring*)realloc(dict,sizeof(t_ANSIstring)*(++dict_sz));
    dict[di].bytes=malloc(ai+1); memcpyb(dict[di].bytes,acumulator,ai+1); dict[di].len=ai+1;

    ai=0;
    if(ai>=acumulator_sz)acumulator=(unsigned char*)realloc(acumulator,++acumulator_sz);
    acumulator[0]=byte;
   }else{
    entry=(long)word;
    entry-=256;
    if(entry>di){
     byte=acumulator[0];

     acumulator[ai]=byte;
     if((++di)>=dict_sz)dict=(t_ANSIstring*)realloc(dict,sizeof(t_ANSIstring)*(++dict_sz));
     dict[di].bytes=malloc(ai+1); memcpyb(dict[di].bytes,acumulator,ai+1); dict[di].len=ai+1;

     write_bytes(out,acumulator,ai+1);
    }else{
      byte=dict[entry].bytes[0];
      write_bytes(out,dict[entry].bytes,dict[entry].len);

      acumulator[ai]=byte;
      if((++di)>=dict_sz)dict=(t_ANSIstring*)realloc(dict,sizeof(t_ANSIstring)*(++dict_sz));
      dict[di].bytes=malloc(ai+1); memcpyb(dict[di].bytes,acumulator,ai+1); dict[di].len=ai+1;

      if(dict[entry].len>acumulator_sz)acumulator=(unsigned char*)realloc(acumulator,acumulator_sz=dict[entry].len);
      memcpyb(acumulator,dict[entry].bytes,ai=dict[entry].len); ai--;
     }
   }
  }else if(word<256){//Esto siempre debería cumplirse en la primera lectura
         byte=(unsigned char)(word&0xFF);
         write_byte(out,byte);

         ai=0;
         if(ai>=acumulator_sz)acumulator=(unsigned char*)realloc(acumulator,++acumulator_sz);
         acumulator[0]=byte;
        }
  word=0;
 }

 flush_write(out);
 if(acumulator)free(acumulator);
 if(dict)free(dict);
}


#define BUFFER_BUFFERED 1024*1024
#ifndef DECOMP
int main(int argc,char **argv){
 if(argc<2)return -1;
 t_stream sin=create_stream(argv[1],O_RDONLY,BUFFER_BUFFERED),
          sout=create_stream(argv[2],O_WRONLY|O_CREAT,BUFFER_BUFFERED);
 printf("Comprimiendo...");
 lz_compress(&sin,&sout);
 printf("Ok\n");
 close_stream(&sin);
 close_stream(&sout);
 return 0;
}
#else
int main(int argc,char **argv){
 if(argc<2)return -1;
 t_stream sin=create_stream(argv[1],O_RDONLY,BUFFER_BUFFERED),
          sout=create_stream(argv[2],O_WRONLY|O_CREAT,BUFFER_BUFFERED);
 printf("Descomprimiendo...");
 lz_expand(&sin,&sout);
 printf("Ok\n");
 close_stream(&sin);
 close_stream(&sout);
 return 0;
}
#endif

