#ifndef __STRUCTURES__
#define __STRUCTURES__
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include "boolean.h"
 typedef struct{
  char Signature[0x13];
  char EOP; // $1A, como el #0.
  unsigned short FirstDataBlockAt __attribute__((packed));
  unsigned short Version __attribute__((packed));
  unsigned short Revision __attribute__((packed));
 }tVOCheader;
/*
 typedef struct{
  unsigned char SampleRate;
  unsigned char CompressionType;
 }tVOCinfo;
*/
 typedef struct{
  unsigned long SamplesPerSecond;
  char BitsPerSample;
  char Channel;
  unsigned short Format;
  char Reserved[4];
 }tVOCinfo;

 typedef struct{
  unsigned short TimeConst __attribute__((packed));
  unsigned char Pack;
  unsigned char Mode;
 }tVOCeXtendedInfo;

 typedef struct{
  tVOCheader Header;
  tVOCinfo Info;
  tVOCeXtendedInfo XInfo;
  unsigned char *Data;
  unsigned int DataLen __attribute__((packed));
  char *Author;
 }tVOC;

 typedef struct{
  unsigned char Blue;
  unsigned char Green;
  unsigned char Red;
  unsigned char Color;
 }tBitmapPalette;

 typedef struct{
  unsigned char ID[2];
  unsigned int FileSize __attribute__((packed));
  unsigned char Reserved[4];
  unsigned int Offset __attribute__((packed));
  unsigned int HeaderSize __attribute__((packed));
  unsigned int Wide __attribute__((packed));
  unsigned int Height __attribute__((packed));
  unsigned short Planes __attribute__((packed));
  unsigned short BitsPerPixel __attribute__((packed));
  unsigned int Compression __attribute__((packed));
  unsigned int ImageSize __attribute__((packed));
  unsigned int HorizontalRes __attribute__((packed));
  unsigned int VerticalRes __attribute__((packed));
  unsigned int UsedColors __attribute__((packed));
  unsigned int ImportantColors __attribute__((packed));
 }tBitmapHeader;

 typedef struct{
  tBitmapPalette *Palette;
  tBitmapHeader Header;
  long Padding;
  char *Data;
 }tBitmap;

 typedef struct{
  unsigned char Blue;
  unsigned char Green;
  unsigned char Red;
  unsigned char Color;
 }tSpritePalette;

 typedef struct{
  unsigned int ColorResolution __attribute__((packed));
  unsigned int Wide __attribute__((packed));
  unsigned int Height __attribute__((packed));
  tSpritePalette *Palette;
  unsigned char *Data;
  bool ApplyTransparent;
  unsigned int TransparentClr; //Para transparencia rapida en esprites.
  unsigned int *Ligth;
  float *Alpha,SingleAlpha;
  unsigned int Rf_x,Rf_y;
 }tSprite;

 typedef struct{
  int x,y;
 }tDiscretePoint;

 typedef struct{
  float x,y,u,v;
 }tTexturePoint;

 typedef struct{
  unsigned char *VRAM8b;
  unsigned short *VRAM16b;
  unsigned int *VRAM32b;
  void *VRAM;
  unsigned short Height,Wide,Deep;
  void *auxDATA;
  char *miscOPTIONS;
  void(*Set)(void*,char**);
  void(*RefreshZone)(void*,int,int,int,int);
  void(*Refresh)(void*);
  void(*Close)(void*);
  char *CardId;
 }tStdVideoFrame;

 typedef struct{
  float x,y;
 }t2DPoint;
 
 typedef struct{
  float x,y,z;
 }t3DPoint;

 typedef void * tAudioHw; //Obsoleta

 typedef struct{
  int rate;
  int mode;
  int sample_sz;
  long len;
  void *data;
 }tSamples;

 typedef struct{
  t3DPoint pos;
  t3DPoint vel;
  t3DPoint at_ang; //Vetores de orientacion  ↑(up)+y
  t3DPoint up_ang; //como en OpenGL        --·-→(at)+z
  float gain;      //Ganancia (dB)           |
 }tAudioListener;

 typedef struct{
  tSamples samples;
  t3DPoint pos;
  t3DPoint vel;
  char follows_listener;
  char play_looping;
 }tAudioSource;

 typedef struct{
  float speed_of_sound;
  float doppler;
  float doppler_factor;
 }tAudioEnviroment;

 typedef struct{
  void *auxDATA;
  char *miscOPTIONS;
  void *PoolData4ISR; //Obsoleto
  unsigned short Hz,Format,Channels,BufferingSize;
  void(*Start)(void*,char**); //Modificada
  void(*Stop)(void*);
  void(*Close)(void*);
  //AÑADIDAS:
  void (*Play)(void*); //tAudioSource
  void (*Pause)(void*); //tAudioSource
  void (*Halt)(void*);
  void (*Set)(void*); //tAudio_Listener
  void*(*CreateListener)(t3DPoint*,t3DPoint*,float,float,float,char**,...);
  void*(*CreateSource)(t3DPoint*,t3DPoint*,char,char);
  void (*DestroyListener)(void*); //tAudio_Listener
  void (*DestroySource)(void*); //tAudioSource
  void*(*CreateEnviroment)(float,float,float);
  void (*DestroyEnviroment)(void*); //tAudioEnviroment
  //
  char *CardId;
 }tStdAudioController;

 typedef struct{
  int meece_x,meece_y,meece_btns,dmeece_x,dmeece_y,kb_scan;
 }tInputStatus;

 typedef struct{
  void *auxDATA;
  char *miscOPTIONS;
  void (*Set)(void*,char**,...);
  void(*KeybClose)(void*);
  long(*KeybAddISR)(void*,void*,char,void*);
  void(*KeybRmISR)(long,char);
  void(*MouseClose)(void*);
  long(*MouseAddISR)(void*,void*,char,void*);
  void(*MouseRmISR)(long,char);
  void(*PadClose)(void*);
  long(*PadAddISR)(void*,char,void*);
  void(*PadRmISR)(long,char);
  void(*Close)(void*);
  char *CardId;
  long (*Poll)(long,tInputStatus*);
  long (*RetPollId)(char*);
 }tStdInput;

 typedef struct{
  void *LibHandler;
  tStdVideoFrame *screen;
  tStdInput *input;
  tStdAudioController *audio;
  char *(*GetCapables)(void);
  int (*IsCapable)(char*);
 }tStdDriverSet;

 typedef struct{
  float x,y,z,u,v,tz,sx,sy;
 }t3DTexPoint;

 typedef struct{
  t3DPoint N; float D;
 }t2DPlane;

 typedef struct{
  t3DPoint P; float r;
 }tCircleOrCube;

 typedef struct{
  unsigned long *ASCIIChar,SymStored; //Aqui se guadan los caracteres ASCII
  tSprite **GraphChar;       //y en la misma posicion relativa, los caracteres tipograficos
  char *Decode;
 }tFont;

 typedef struct{
  tFont *Font;
  long Fonts;
 }tFontSet;

 typedef struct{
  long Entries;
  char **Entry;
 }tStdGetDirContents;

 typedef struct{
  float x,y,z,DistanceXYZ;
 }t3DOffset;

 typedef struct{
  float x,y,z,u,v,DistanceXYZ,DistanceUV;
 }t3DTexOffset;

 typedef struct{
  long StringSz;
  long ChrSz;
  void *String;
 }tIndexString;

 typedef struct{
  long SleepTicks,CurrentTicks;
  float SleepSeconds;
 }tGCclock;

 typedef struct{
  struct dirent Entry;
  void *Next;
 }tDirEntryChain;

 typedef struct{
  void *user_data;
  char *wall_name;
  void *Next;
 }tWallTail;

 typedef struct{
  void*(*insert)(void*,void*);
  void*(*unsafe_insert)(void*,void*);
  void(*delete_item)(void*,void*);
  void(*delete_data)(void*,void*);
  void(*dispose)(void*);
  void(*clear)(void*);
  void*(*find)(void*,void*);
  void*(*find_from)(void*,void*);
  void**(*dump)(void*);
  unsigned long(*count)(void*);
 }tADTfunctions;

 typedef struct{
  tADTfunctions *f;
  void *handler;
 }tADT;

 typedef struct{
  tStdDriverSet *driver;
  tInputStatus input;
 }tArtifactSystemData;

 typedef struct{
  long Type;
  void*(*Create)(void*);
  int(*Destroy)(void*);
  void*(*Refresh)(void*);
  char *Name;
  tArtifactSystemData *SystemData;
  void *UserData,*Next,*Prev;
  int Enabled;
 }tArtifact;

 typedef struct{
  void *LibHndlr;
  tArtifact*(*create_builder)(tStdDriverSet*,char);
  void(*destroy_builder)(tArtifact*);
  tArtifact*(*build_artifact)(tArtifact*,long,long,char*);
  tArtifact*(*build_artifact_by_name)(tArtifact*,char*,char*,char*);
  void(*destroy_artifact)(tArtifact*);
  int(*set_field)(tArtifact*,char*,...);
  int(*get_field)(tArtifact*,char*,...);
  tArtifact*(*RetArtByName)(tArtifact*,char*);
  long(*return_sym)(char*);
  char *Name;
 }tArtifactLib;

 typedef struct{
  tSprite *This; char *Name;
  void *Next,*Prev;
 }tSpriteListEntry;

 typedef struct{
  tFont *This; char *Name;
  void *Next,*Prev;
 }tFontListEntry;

 typedef struct{
  char *Caption; void *Next,*Prev;
 }tCharListEntry;

 typedef struct{
  unsigned long Artifacts,Libs;
  tArtifact **Artifact;
  tArtifactLib *Lib;
  tSpriteListEntry *Image;
  tFontListEntry *Font;
  tCharListEntry *Caption;
 }tArtifactSet;

 typedef struct{
  void *so_handler,*session;
  char *diag;

  void(*CloseSession)(void*);
  int(*OpenSession)(int,int,int,void*,void*);
  void*(*AddTexture)(void*,const tSprite*);//-
  void(*DeleteTexture)(void*,void*);//-
  void(*SetBackground)(void*,void*);//-
  void*(*AddBackground)(void*,void*);//-
  void(*Move3DItem)(void*,void*,t3DPoint*,t3DPoint*);
  void(*Rotate3DItem)(void*,void*,t3DPoint*);
  void(*LimitMove3DItem[4])(void*,void*,t3DPoint*,t3DPoint*,void**);
  void(*Move3DItemOFF)(void*,void*,t3DPoint*);
  void(*LimitMove3DItemOFF[2])(void*,void*,t3DPoint*,void**);
  void(*CtrlMvItem)(void*,void*,void*,t3DPoint*,t3DPoint*,char**);
  void(*CtrlMvItemOFF)(void*,void*,void*,t3DPoint*,char**);
  void(*ListLimitMove3DItemOFF)(void*,void*,t3DPoint*,tWallTail*);
  void(*ListLimitMove3DItem)(void*,void*,t3DPoint*,t3DPoint*,tWallTail*);
  void(*ItemIntersection)(void*,void*,tWallTail*);
  void(*ListTraceMove3DItem)(void*,void*,t3DPoint*,tWallTail*);
  void(*Walk3D)(void*,t3DPoint*,t3DPoint*);
  void(*FlushWalk3D)(void*);
  void(*ModifyViewBox)(void*,float,float,float,float);
  void(*FlushView)(void*);
  void(*ClearZbuffer)(void*);
  void(*Draw3DWorld)(void*);
  void*(*FullInsert3DItem)(void*,void*,char*);
  void(*Insert3DItem)(void*,void*);
  void(*InsertControlTriangle)(void*,void*,char);
  void(*RemoveItem)(void*);
  void(*FullRemoveItem)(void*,void*);
  void*(*Retry3DItemEntry)(char*,void*);
  void*(*Retry3DItem)(char*,void*);
  void*(*Retry3DWallEntry)(char*,void*);
  void*(*FullInsert3DWall)(void*,void*,char*);
  void(*FlipY3DItem)(void*);
  void(*FlipY3DItemCC)(void*);
  void(*Hide3DItem)(void*);
  void(*Show3DItem)(void*);
  void*(*Find3DItemListEntry)(char*,void*);

  void*(*Load3Dres)(void*,char,char*,char**);
  void(*Release3Dres)(void*);//-
  void(*Save3Dres)(void*,char*,char**);//-
  void(*SetActive3Dres)(void*,void*);

  int(*GetCtrlBits)(char*);
  void(*Copy)(void*,int,void*);

  //void*(*GetAddrMap)(int,void*); //Retirada en la version 9.0.0
  void*(*GetAddrMapOff)(void*,int,int);   //}/__Añadidas en la
  void*(*GetAddrMapXY)(void*,int,int,int);//}\  version 9.0.0.
  void(*ChangeItemUserData)(void*,void*);
  t3DPoint*(*GetItemHold)(void*);

  void*(*GetTriangleUserData)(void*);
  void*(*GetItemUserData)(void*);
  char*(*GetItemName)(void*);
  char*(*GetItemNameFromAddrMap)(void*);
  char*(*GetTriangleName)(void*);
  char*(*GetTriangleNameFromAddrMap)(void*);
 }t3DW_f;

#endif


